@@ -1,19 +1,62 @@
+//! Implements Dijkstra's algorithm for shortest path, to find an
+//! optimal and readable diff between two ASTs.
+
 use std::cmp::{min, Ordering, Reverse};
 use std::collections::BinaryHeap;
 use std::hash::{Hash, Hasher};
 
-use crate::lines::LineNumber;
 use crate::syntax::{ChangeKind, Syntax};
 use rustc_hash::FxHashMap;
 use strsim::normalized_levenshtein;
 use Edge::*;
 
+/// A vertex in a directed acyclic graph that represents a diff.
+///
+/// Each vertex represents two pointers: one to the next unmatched LHS
+/// syntax, and one to the next unmatched RHS syntax.
+///
+/// For example, suppose we have `X A` on the LHS and `A` on the
+/// RHS. Our start vertex looks like this.
+///
+/// ```
+/// LHS: X A     RHS: A
+///      ^            ^
+/// ```
+///
+/// From this vertex, we could take [`Edge::NovelAtomLHS`], bringing
+/// us to this vertex.
+///
+/// ```
+/// LHS: X A     RHS: A
+///        ^          ^
+/// ```
+///
+/// Alternatively, we could take the [`Edge::NovelAtomRHS`], bringing us
+/// to this vertex.
+///
+/// ```
+/// LHS: X A     RHS: A
+///      ^              ^
+/// ```
 #[derive(Debug, Clone)]
 struct Vertex<'a> {
     lhs_syntax: Option<&'a Syntax<'a>>,
-    lhs_prev_novel: Option<LineNumber>,
     rhs_syntax: Option<&'a Syntax<'a>>,
-    rhs_prev_novel: Option<LineNumber>,
+    /// If the previous edge was marking syntax as novel, what line
+    /// was it on?
+    ///
+    /// We want to prefer marking syntax nodes as novel if we've
+    /// already marked other nodes as novel on the current line. See
+    /// `sample_files/contiguous_after.js`.
+    ///
+    /// Unfortunately this is path dependent: the vertex doesn't care
+    /// how we got here.
+    ///
+    /// We solve this by distinguishing vertices based on the novel
+    /// state of the previous edge. This increases the search space
+    /// but allows us to keep using a graph traversal.
+    lhs_prev_is_novel: bool,
+    rhs_prev_is_novel: bool,
 }
 impl<'a> PartialEq for Vertex<'a> {
     fn eq(&self, other: &Self) -> bool {
@@ -36,18 +79,26 @@ impl<'a> Vertex<'a> {
     }
 }
 
-// Rust requires that PartialEq, PartialOrd and Ord agree.
-// https://doc.rust-lang.org/std/cmp/trait.Ord.html
-//
-// For `Vertex`, we want to compare by distance in a priority queue, but
-// equality should only consider LHS/RHS node when deciding if we've
-// visited a vertex. We define separate wrappers for these two use
-// cases.
+/// A vertex with a distance.
+///
+/// Rust requires that PartialEq, PartialOrd and Ord agree.
+/// <https://doc.rust-lang.org/std/cmp/trait.Ord.html>
+///
+/// We want two nodes of the same distance to be the same as far as
+/// the priority queue is concerned. This differs from the equality
+/// semantics on [`Vertex`] (which should only consider LHS/RHS
+/// syntax).
+///
+/// Defining OrdVertex as a separate type allows us different Eq and
+/// Ord behaviour.
 #[derive(Debug)]
 struct OrdVertex<'a> {
+    /// The distance from the start vertex to this vertex.
     distance: u64,
+    /// The vertex we have just reached.
+    current: Vertex<'a>,
+    /// The previous vertex and edge we took to reach this vertex.
     prev: Option<(Vertex<'a>, Edge)>,
-    v: Vertex<'a>,
 }
 
 impl<'a> PartialOrd for OrdVertex<'a> {
@@ -69,11 +120,18 @@ impl<'a> PartialEq for OrdVertex<'a> {
 }
 impl<'a> Eq for OrdVertex<'a> {}
 
+/// An edge in our graph, with an associated [`cost`](Edge::cost).
+///
+/// A syntax node can always be marked as novel, so a vertex will have
+/// at least a NovelFoo edge. Depending on the syntax nodes of the
+/// current [`Vertex`], other edges may also be available.
+///
+/// See [`neighbours`] for all the edges available for a given `Vertex`.
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
 enum Edge {
-    UnchangedNode(u64),
-    UnchangedDelimiter(u64),
-    ReplacedComment,
+    UnchangedNode { depth_difference: u64 },
+    UnchangedDelimiter { depth_difference: u64 },
+    ReplacedComment { levenshtein_pct: u8 },
     NovelAtomLHS { contiguous: bool },
     NovelAtomRHS { contiguous: bool },
     NovelDelimiterLHS { contiguous: bool },
@@ -86,26 +144,28 @@ impl Edge {
     fn cost(&self) -> u64 {
         match self {
             // Matching nodes is always best.
-            UnchangedNode(depth_difference) => min(40, *depth_difference),
+            UnchangedNode { depth_difference } => min(40, *depth_difference),
             // Matching an outer delimiter is good.
-            UnchangedDelimiter(depth_difference) => 100 + min(40, *depth_difference),
+            UnchangedDelimiter { depth_difference } => 100 + min(40, *depth_difference),
 
             // Replacing a comment is better than treating it as novel.
-            ReplacedComment => 150,
+            ReplacedComment { levenshtein_pct } => 150 + u64::from(100 - levenshtein_pct),
 
             // Otherwise, we've added/removed a node.
-            NovelAtomLHS { contiguous } | NovelAtomRHS { contiguous } => {
+            NovelAtomLHS { contiguous }
+            | NovelAtomRHS { contiguous }
+            | NovelDelimiterLHS { contiguous }
+            | NovelDelimiterRHS { contiguous } => {
                 if *contiguous {
-                    200
+                    300
                 } else {
-                    201
-                }
-            }
-            NovelDelimiterLHS { contiguous } | NovelDelimiterRHS { contiguous } => {
-                if *contiguous {
-                    200
-                } else {
-                    201
+                    // This needs to be more than 40 greater than the
+                    // contiguous case. Otherwise, we end up choosing
+                    // a badly positioned unchanged elimiter just
+                    // because it has a better depth difference.
+                    //
+                    // TODO: write a test for this case.
+                    350
                 }
             }
 
@@ -113,7 +173,7 @@ impl Edge {
             // novel rather than marking 90% of the children as
             // novel. This stops us matching up completely unrelated trees.
             NovelTreeLHS { num_descendants } | NovelTreeRHS { num_descendants } => {
-                200 + (*num_descendants - 10) * NovelDelimiterLHS { contiguous: false }.cost()
+                300 + (*num_descendants - 10) * NovelDelimiterLHS { contiguous: false }.cost()
             }
         }
     }
@@ -127,8 +187,8 @@ fn shortest_path(start: Vertex) -> Vec<(Edge, Vertex)> {
 
     heap.push(Reverse(OrdVertex {
         distance: 0,
+        current: start,
         prev: None,
-        v: start.clone(),
     }));
 
     // TODO: this grows very big. Consider using IDA* to reduce memory
@@ -138,27 +198,31 @@ fn shortest_path(start: Vertex) -> Vec<(Edge, Vertex)> {
     let end;
     loop {
         match heap.pop() {
-            Some(Reverse(OrdVertex { distance, prev, v })) => {
-                if predecessors.contains_key(&v) {
+            Some(Reverse(OrdVertex {
+                distance,
+                current,
+                prev,
+            })) => {
+                if predecessors.contains_key(&current) {
                     continue;
                 }
-                predecessors.insert(v.clone(), prev);
+                predecessors.insert(current.clone(), prev);
 
-                if v.is_end() {
-                    end = v;
+                if current.is_end() {
+                    end = current;
                     break;
                 }
 
-                for (edge, new_v) in neighbours(&v) {
-                    if predecessors.contains_key(&new_v) {
+                for (edge, next) in neighbours(&current) {
+                    if predecessors.contains_key(&next) {
                         continue;
                     }
-                    let new_v_distance = distance + edge.cost();
+                    let distance_to_next = distance + edge.cost();
 
                     heap.push(Reverse(OrdVertex {
-                        distance: new_v_distance,
-                        prev: Some((v.clone(), edge)),
-                        v: new_v,
+                        distance: distance_to_next,
+                        current: next,
+                        prev: Some((current.clone(), edge)),
                     }));
                 }
             }
@@ -184,18 +248,18 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
 
     if let (Some(lhs_syntax), Some(rhs_syntax)) = (&v.lhs_syntax, &v.rhs_syntax) {
         if lhs_syntax.equal_content(rhs_syntax) {
-            let depth_difference = (lhs_syntax.info().num_ancestors.get() as i64
-                - rhs_syntax.info().num_ancestors.get() as i64)
+            let depth_difference = (lhs_syntax.num_ancestors() as i64
+                - rhs_syntax.num_ancestors() as i64)
                 .abs() as u64;
 
             // Both nodes are equal, the happy case.
             res.push((
-                UnchangedNode(depth_difference),
+                UnchangedNode { depth_difference },
                 Vertex {
                     lhs_syntax: lhs_syntax.next(),
                     rhs_syntax: rhs_syntax.next(),
-                    lhs_prev_novel: None,
-                    rhs_prev_novel: None,
+                    lhs_prev_is_novel: false,
+                    rhs_prev_is_novel: false,
                 },
             ));
         }
@@ -228,17 +292,17 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
                     Some(rhs_children[0])
                 };
 
-                let depth_difference = (lhs_syntax.info().num_ancestors.get() as i64
-                    - rhs_syntax.info().num_ancestors.get() as i64)
+                let depth_difference = (lhs_syntax.num_ancestors() as i64
+                    - rhs_syntax.num_ancestors() as i64)
                     .abs() as u64;
 
                 res.push((
-                    UnchangedDelimiter(depth_difference),
+                    UnchangedDelimiter { depth_difference },
                     Vertex {
                         lhs_syntax: lhs_next,
-                        lhs_prev_novel: None,
                         rhs_syntax: rhs_next,
-                        rhs_prev_novel: None,
+                        lhs_prev_is_novel: false,
+                        rhs_prev_is_novel: false,
                     },
                 ));
             }
@@ -259,17 +323,16 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
         {
             // Both sides are comments and their content is reasonably
             // similar.
-            if *lhs_is_comment
-                && *rhs_is_comment
-                && normalized_levenshtein(lhs_content, rhs_content) > 0.4
-            {
+            if *lhs_is_comment && *rhs_is_comment && lhs_content != rhs_content {
+                let levenshtein_pct =
+                    (normalized_levenshtein(lhs_content, rhs_content) * 100.0).round() as u8;
                 res.push((
-                    ReplacedComment,
+                    ReplacedComment { levenshtein_pct },
                     Vertex {
                         lhs_syntax: lhs_syntax.next(),
-                        lhs_prev_novel: None,
                         rhs_syntax: rhs_syntax.next(),
-                        rhs_prev_novel: None,
+                        lhs_prev_is_novel: false,
+                        rhs_prev_is_novel: false,
                     },
                 ));
             }
@@ -282,19 +345,18 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
             Syntax::Atom { .. } => {
                 res.push((
                     NovelAtomLHS {
-                        contiguous: v.lhs_prev_novel == lhs_syntax.first_line(),
+                        contiguous: v.lhs_prev_is_novel && lhs_syntax.prev_is_contiguous(),
                     },
                     Vertex {
                         lhs_syntax: lhs_syntax.next(),
-                        lhs_prev_novel: lhs_syntax.last_line(),
                         rhs_syntax: v.rhs_syntax,
-                        rhs_prev_novel: v.rhs_prev_novel,
+                        lhs_prev_is_novel: true,
+                        rhs_prev_is_novel: v.rhs_prev_is_novel,
                     },
                 ));
             }
             // Step into this partially/fully novel list.
             Syntax::List {
-                open_position,
                 children,
                 num_descendants,
                 ..
@@ -302,19 +364,18 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
                 let lhs_next = if children.is_empty() {
                     lhs_syntax.next()
                 } else {
-                    // `lhs_prev_novel` only tracks nodes at the same level.
                     Some(children[0])
                 };
 
                 res.push((
                     NovelDelimiterLHS {
-                        contiguous: v.lhs_prev_novel == lhs_syntax.first_line(),
+                        contiguous: v.lhs_prev_is_novel && lhs_syntax.prev_is_contiguous(),
                     },
                     Vertex {
                         lhs_syntax: lhs_next,
-                        lhs_prev_novel: open_position.last().map(|lp| lp.line),
                         rhs_syntax: v.rhs_syntax,
-                        rhs_prev_novel: v.rhs_prev_novel,
+                        lhs_prev_is_novel: true,
+                        rhs_prev_is_novel: v.rhs_prev_is_novel,
                     },
                 ));
 
@@ -325,9 +386,9 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
                         },
                         Vertex {
                             lhs_syntax: lhs_syntax.next(),
-                            lhs_prev_novel: v.lhs_prev_novel,
                             rhs_syntax: v.rhs_syntax,
-                            rhs_prev_novel: v.rhs_prev_novel,
+                            lhs_prev_is_novel: v.lhs_prev_is_novel,
+                            rhs_prev_is_novel: v.rhs_prev_is_novel,
                         },
                     ));
                 }
@@ -341,19 +402,18 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
             Syntax::Atom { .. } => {
                 res.push((
                     NovelAtomRHS {
-                        contiguous: v.rhs_prev_novel == rhs_syntax.first_line(),
+                        contiguous: v.rhs_prev_is_novel && rhs_syntax.prev_is_contiguous(),
                     },
                     Vertex {
                         lhs_syntax: v.lhs_syntax,
-                        lhs_prev_novel: v.lhs_prev_novel,
                         rhs_syntax: rhs_syntax.next(),
-                        rhs_prev_novel: rhs_syntax.last_line(),
+                        lhs_prev_is_novel: v.lhs_prev_is_novel,
+                        rhs_prev_is_novel: true,
                     },
                 ));
             }
             // Step into this partially/fully novel list.
             Syntax::List {
-                open_position,
                 children,
                 num_descendants,
                 ..
@@ -366,13 +426,13 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
 
                 res.push((
                     NovelDelimiterRHS {
-                        contiguous: v.rhs_prev_novel == rhs_syntax.first_line(),
+                        contiguous: v.rhs_prev_is_novel && rhs_syntax.prev_is_contiguous(),
                     },
                     Vertex {
                         lhs_syntax: v.lhs_syntax,
-                        lhs_prev_novel: v.lhs_prev_novel,
                         rhs_syntax: rhs_next,
-                        rhs_prev_novel: open_position.last().map(|lp| lp.line),
+                        lhs_prev_is_novel: v.lhs_prev_is_novel,
+                        rhs_prev_is_novel: true,
                     },
                 ));
 
@@ -383,9 +443,9 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
                         },
                         Vertex {
                             lhs_syntax: v.lhs_syntax,
-                            lhs_prev_novel: v.lhs_prev_novel,
                             rhs_syntax: rhs_syntax.next(),
-                            rhs_prev_novel: v.rhs_prev_novel,
+                            lhs_prev_is_novel: v.lhs_prev_is_novel,
+                            rhs_prev_is_novel: v.rhs_prev_is_novel,
                         },
                     ));
                 }
@@ -399,9 +459,9 @@ fn neighbours<'a>(v: &Vertex<'a>) -> Vec<(Edge, Vertex<'a>)> {
 pub fn mark_syntax<'a>(lhs_syntax: Option<&'a Syntax<'a>>, rhs_syntax: Option<&'a Syntax<'a>>) {
     let start = Vertex {
         lhs_syntax,
-        lhs_prev_novel: None,
         rhs_syntax,
-        rhs_prev_novel: None,
+        lhs_prev_is_novel: false,
+        rhs_prev_is_novel: false,
     };
     let route = shortest_path(start);
     mark_route(&route);
@@ -410,14 +470,14 @@ pub fn mark_syntax<'a>(lhs_syntax: Option<&'a Syntax<'a>>, rhs_syntax: Option<&'
 fn mark_route(route: &[(Edge, Vertex)]) {
     for (e, v) in route {
         match e {
-            UnchangedNode(_) => {
+            UnchangedNode { .. } => {
                 // No change on this node or its children.
                 let lhs = v.lhs_syntax.unwrap();
                 let rhs = v.rhs_syntax.unwrap();
                 lhs.set_change_deep(ChangeKind::Unchanged(rhs));
                 rhs.set_change_deep(ChangeKind::Unchanged(lhs));
             }
-            UnchangedDelimiter(_) => {
+            UnchangedDelimiter { .. } => {
                 // No change on the outer delimiter, but children may
                 // have changed.
                 let lhs = v.lhs_syntax.unwrap();
@@ -425,11 +485,17 @@ fn mark_route(route: &[(Edge, Vertex)]) {
                 lhs.set_change(ChangeKind::Unchanged(rhs));
                 rhs.set_change(ChangeKind::Unchanged(lhs));
             }
-            ReplacedComment => {
+            ReplacedComment { levenshtein_pct } => {
                 let lhs = v.lhs_syntax.unwrap();
                 let rhs = v.rhs_syntax.unwrap();
-                lhs.set_change(ChangeKind::ReplacedComment(lhs, rhs));
-                rhs.set_change(ChangeKind::ReplacedComment(rhs, lhs));
+
+                if *levenshtein_pct > 40 {
+                    lhs.set_change(ChangeKind::ReplacedComment(lhs, rhs));
+                    rhs.set_change(ChangeKind::ReplacedComment(rhs, lhs));
+                } else {
+                    lhs.set_change(ChangeKind::Novel);
+                    rhs.set_change(ChangeKind::Novel);
+                }
             }
             NovelAtomLHS { .. } | NovelDelimiterLHS { .. } => {
                 let lhs = v.lhs_syntax.unwrap();
@@ -506,52 +572,62 @@ mod tests {
 
         let start = Vertex {
             lhs_syntax: Some(lhs),
-            lhs_prev_novel: None,
             rhs_syntax: Some(rhs),
-            rhs_prev_novel: None,
-        };
-        let route = shortest_path(start);
-
-        let actions = route.iter().map(|(action, _)| *action).collect_vec();
-        assert_eq!(actions, vec![UnchangedNode(0)]);
-    }
-
-    #[test]
-    fn extra_atom_lhs() {
-        let arena = Arena::new();
-
-        let lhs: Vec<&Syntax> = vec![Syntax::new_list(
-            &arena,
-            "[".into(),
-            pos_helper(0),
-            vec![Syntax::new_atom(&arena, pos_helper(1), "foo")],
-            "]".into(),
-            pos_helper(2),
-        )];
-        init_info(&lhs);
-
-        let rhs: Vec<&Syntax> = vec![Syntax::new_list(
-            &arena,
-            "[".into(),
-            pos_helper(0),
-            vec![],
-            "]".into(),
-            pos_helper(2),
-        )];
-        init_info(&rhs);
-
-        let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
         let actions = route.iter().map(|(action, _)| *action).collect_vec();
         assert_eq!(
             actions,
-            vec![UnchangedDelimiter(0), NovelAtomLHS { contiguous: false }]
+            vec![UnchangedNode {
+                depth_difference: 0
+            }]
+        );
+    }
+
+    #[test]
+    fn extra_atom_lhs() {
+        let arena = Arena::new();
+
+        let lhs = vec![Syntax::new_list(
+            &arena,
+            "[",
+            pos_helper(0),
+            vec![Syntax::new_atom(&arena, pos_helper(1), "foo")],
+            "]",
+            pos_helper(2),
+        )];
+        init_info(&lhs);
+
+        let rhs = vec![Syntax::new_list(
+            &arena,
+            "[",
+            pos_helper(0),
+            vec![],
+            "]",
+            pos_helper(2),
+        )];
+        init_info(&rhs);
+
+        let start = Vertex {
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
+        };
+        let route = shortest_path(start);
+
+        let actions = route.iter().map(|(action, _)| *action).collect_vec();
+        assert_eq!(
+            actions,
+            vec![
+                UnchangedDelimiter {
+                    depth_difference: 0
+                },
+                NovelAtomLHS { contiguous: false }
+            ]
         );
     }
 
@@ -559,34 +635,34 @@ mod tests {
     fn repeated_atoms() {
         let arena = Arena::new();
 
-        let lhs: Vec<&Syntax> = vec![Syntax::new_list(
+        let lhs = vec![Syntax::new_list(
             &arena,
-            "[".into(),
+            "[",
             pos_helper(0),
             vec![],
-            "]".into(),
+            "]",
             pos_helper(2),
         )];
         init_info(&lhs);
 
-        let rhs: Vec<&Syntax> = vec![Syntax::new_list(
+        let rhs = vec![Syntax::new_list(
             &arena,
-            "[".into(),
+            "[",
             pos_helper(0),
             vec![
                 Syntax::new_atom(&arena, pos_helper(1), "foo"),
                 Syntax::new_atom(&arena, pos_helper(2), "foo"),
             ],
-            "]".into(),
+            "]",
             pos_helper(3),
         )];
         init_info(&rhs);
 
         let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
@@ -594,7 +670,9 @@ mod tests {
         assert_eq!(
             actions,
             vec![
-                UnchangedDelimiter(0),
+                UnchangedDelimiter {
+                    depth_difference: 0
+                },
                 NovelAtomRHS { contiguous: false },
                 NovelAtomRHS { contiguous: false }
             ]
@@ -605,51 +683,37 @@ mod tests {
     fn atom_after_empty_list() {
         let arena = Arena::new();
 
-        let lhs: Vec<&Syntax> = vec![Syntax::new_list(
+        let lhs = vec![Syntax::new_list(
             &arena,
-            "[".into(),
+            "[",
             pos_helper(0),
             vec![
-                Syntax::new_list(
-                    &arena,
-                    "(".into(),
-                    pos_helper(1),
-                    vec![],
-                    ")".into(),
-                    pos_helper(2),
-                ),
+                Syntax::new_list(&arena, "(", pos_helper(1), vec![], ")", pos_helper(2)),
                 Syntax::new_atom(&arena, pos_helper(3), "foo"),
             ],
-            "]".into(),
+            "]",
             pos_helper(4),
         )];
         init_info(&lhs);
 
-        let rhs: Vec<&Syntax> = vec![Syntax::new_list(
+        let rhs = vec![Syntax::new_list(
             &arena,
-            "{".into(),
+            "{",
             pos_helper(0),
             vec![
-                Syntax::new_list(
-                    &arena,
-                    "(".into(),
-                    pos_helper(1),
-                    vec![],
-                    ")".into(),
-                    pos_helper(2),
-                ),
+                Syntax::new_list(&arena, "(", pos_helper(1), vec![], ")", pos_helper(2)),
                 Syntax::new_atom(&arena, pos_helper(3), "foo"),
             ],
-            "}".into(),
+            "}",
             pos_helper(4),
         )];
         init_info(&rhs);
 
         let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
@@ -659,8 +723,12 @@ mod tests {
             vec![
                 NovelDelimiterRHS { contiguous: false },
                 NovelDelimiterLHS { contiguous: false },
-                UnchangedNode(0),
-                UnchangedNode(0)
+                UnchangedNode {
+                    depth_difference: 0
+                },
+                UnchangedNode {
+                    depth_difference: 0
+                }
             ],
         );
     }
@@ -669,21 +737,21 @@ mod tests {
     fn prefer_atoms_same_line() {
         let arena = Arena::new();
 
-        let lhs: Vec<&Syntax> = vec![
+        let lhs = vec![
             Syntax::new_atom(&arena, col_helper(1, 0), "foo"),
             Syntax::new_atom(&arena, col_helper(2, 0), "bar"),
             Syntax::new_atom(&arena, col_helper(2, 1), "foo"),
         ];
         init_info(&lhs);
 
-        let rhs: Vec<&Syntax> = vec![Syntax::new_atom(&arena, col_helper(1, 0), "foo")];
+        let rhs = vec![Syntax::new_atom(&arena, col_helper(1, 0), "foo")];
         init_info(&rhs);
 
         let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
@@ -691,7 +759,9 @@ mod tests {
         assert_eq!(
             actions,
             vec![
-                UnchangedNode(0),
+                UnchangedNode {
+                    depth_difference: 0
+                },
                 NovelAtomLHS { contiguous: false },
                 NovelAtomLHS { contiguous: true },
             ]
@@ -702,23 +772,23 @@ mod tests {
     fn prefer_children_same_line() {
         let arena = Arena::new();
 
-        let lhs: Vec<&Syntax> = vec![Syntax::new_list(
+        let lhs = vec![Syntax::new_list(
             &arena,
-            "[".into(),
+            "[",
             col_helper(1, 0),
             vec![Syntax::new_atom(&arena, col_helper(1, 2), "1")],
-            "]".into(),
+            "]",
             pos_helper(2),
         )];
         init_info(&lhs);
 
-        let rhs: Vec<&Syntax> = vec![];
+        let rhs = vec![];
 
         let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
@@ -732,13 +802,51 @@ mod tests {
         );
     }
 
+    #[test]
+    fn atom_after_novel_list_contiguous() {
+        let arena = Arena::new();
+
+        let lhs = vec![
+            Syntax::new_list(
+                &arena,
+                "[",
+                col_helper(1, 0),
+                vec![Syntax::new_atom(&arena, col_helper(1, 2), "1")],
+                "]",
+                col_helper(2, 1),
+            ),
+            Syntax::new_atom(&arena, col_helper(2, 2), ";"),
+        ];
+        init_info(&lhs);
+
+        let rhs = vec![];
+
+        let start = Vertex {
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
+        };
+        let route = shortest_path(start);
+
+        let actions = route.iter().map(|(action, _)| *action).collect_vec();
+        assert_eq!(
+            actions,
+            vec![
+                NovelDelimiterLHS { contiguous: false },
+                NovelAtomLHS { contiguous: true },
+                NovelAtomLHS { contiguous: true }
+            ]
+        );
+    }
+
     #[test]
     fn test_novel_tree() {
         let arena = Arena::new();
 
-        let lhs: Vec<&Syntax> = vec![Syntax::new_list(
+        let lhs = vec![Syntax::new_list(
             &arena,
-            "[".into(),
+            "[",
             pos_helper(0),
             vec![
                 Syntax::new_atom(&arena, pos_helper(1), "1"),
@@ -763,14 +871,14 @@ mod tests {
                 Syntax::new_atom(&arena, pos_helper(20), "20"),
                 Syntax::new_atom(&arena, pos_helper(21), "21"),
             ],
-            "]".into(),
+            "]",
             pos_helper(100),
         )];
         init_info(&lhs);
 
-        let rhs: Vec<&Syntax> = vec![Syntax::new_list(
+        let rhs = vec![Syntax::new_list(
             &arena,
-            "[".into(),
+            "[",
             pos_helper(0),
             vec![
                 Syntax::new_atom(&arena, pos_helper(1), "d1"),
@@ -796,16 +904,16 @@ mod tests {
                 Syntax::new_atom(&arena, pos_helper(20), "d20"),
                 Syntax::new_atom(&arena, pos_helper(21), "d21"),
             ],
-            "]".into(),
+            "]",
             pos_helper(100),
         )];
         init_info(&rhs);
 
         let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
@@ -822,18 +930,19 @@ mod tests {
             ]
         );
     }
+
     #[test]
     fn replace_similar_comment() {
         let arena = Arena::new();
 
-        let lhs: Vec<&Syntax> = vec![Syntax::new_comment(
+        let lhs = vec![Syntax::new_comment(
             &arena,
             pos_helper(1),
             "the quick brown fox",
         )];
         init_info(&lhs);
 
-        let rhs: Vec<&Syntax> = vec![Syntax::new_comment(
+        let rhs = vec![Syntax::new_comment(
             &arena,
             pos_helper(1),
             "the quick brown cat",
@@ -841,36 +950,75 @@ mod tests {
         init_info(&rhs);
 
         let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
         let actions = route.iter().map(|(action, _)| *action).collect_vec();
-        assert_eq!(actions, vec![ReplacedComment]);
+        assert_eq!(
+            actions,
+            vec![ReplacedComment {
+                levenshtein_pct: 84
+            }]
+        );
     }
 
     #[test]
     fn replace_very_different_comment() {
         let arena = Arena::new();
 
-        let lhs: Vec<&Syntax> = vec![Syntax::new_comment(
+        let lhs = vec![Syntax::new_comment(
             &arena,
             pos_helper(1),
             "the quick brown fox",
         )];
         init_info(&lhs);
 
-        let rhs: Vec<&Syntax> = vec![Syntax::new_comment(&arena, pos_helper(1), "foo bar")];
+        let rhs = vec![Syntax::new_comment(&arena, pos_helper(1), "foo bar")];
         init_info(&rhs);
 
         let start = Vertex {
-            lhs_syntax: lhs.get(0).map(|n| *n),
-            lhs_prev_novel: None,
-            rhs_syntax: rhs.get(0).map(|n| *n),
-            rhs_prev_novel: None,
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
+        };
+        let route = shortest_path(start);
+
+        let actions = route.iter().map(|(action, _)| *action).collect_vec();
+        assert_eq!(
+            actions,
+            vec![ReplacedComment {
+                levenshtein_pct: 11
+            }]
+        );
+    }
+
+    #[test]
+    fn replace_comment_prefer_most_similar() {
+        let arena = Arena::new();
+
+        let lhs = vec![
+            Syntax::new_comment(&arena, pos_helper(1), "the quick brown fox"),
+            Syntax::new_comment(&arena, pos_helper(2), "the quick brown thing"),
+        ];
+        init_info(&lhs);
+
+        let rhs = vec![Syntax::new_comment(
+            &arena,
+            pos_helper(1),
+            "the quick brown fox.",
+        )];
+        init_info(&rhs);
+
+        let start = Vertex {
+            lhs_syntax: lhs.get(0).copied(),
+            rhs_syntax: rhs.get(0).copied(),
+            lhs_prev_is_novel: false,
+            rhs_prev_is_novel: false,
         };
         let route = shortest_path(start);
 
@@ -878,8 +1026,10 @@ mod tests {
         assert_eq!(
             actions,
             vec![
-                NovelAtomLHS { contiguous: false },
-                NovelAtomRHS { contiguous: false }
+                ReplacedComment {
+                    levenshtein_pct: 95
+                },
+                NovelAtomLHS { contiguous: false }
             ]
         );
     }
