@@ -1,19 +1,17 @@
+//! Manipulate lines of text and groups of lines.
+
 use crate::intervals::Interval;
 use crate::positions::SingleLineSpan;
-use crate::syntax::{aligned_lines, MatchKind, MatchedPos};
+use crate::syntax::MatchedPos;
 use lazy_static::lazy_static;
 use regex::Regex;
 use std::cmp::{max, min, Ordering};
-use std::collections::HashMap;
 use std::fmt;
 
-const SPACER: &str = "  ";
 const MAX_GAP: usize = 1;
-const MIN_WIDTH: usize = 35;
-
-#[cfg(test)]
-use pretty_assertions::assert_eq;
 
+/// A distinct number type for line numbers, to prevent confusion with
+/// other numerical data.
 #[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub struct LineNumber(pub usize);
 
@@ -29,10 +27,13 @@ impl From<usize> for LineNumber {
     }
 }
 
+/// A group of lines that are displayed together.
+///
+/// This is called a 'hunk' in some diff tools.
 #[derive(Debug, PartialEq, Eq)]
 pub struct LineGroup {
-    lhs_lines: Option<Interval<LineNumber>>,
-    rhs_lines: Option<Interval<LineNumber>>,
+    pub lhs_lines: Option<Interval<LineNumber>>,
+    pub rhs_lines: Option<Interval<LineNumber>>,
 }
 
 impl LineGroup {
@@ -43,7 +44,7 @@ impl LineGroup {
         }
     }
 
-    fn lhs_lines(&self) -> Vec<LineNumber> {
+    pub fn lhs_lines(&self) -> Vec<LineNumber> {
         let mut res = vec![];
         match &self.lhs_lines {
             Some(lhs_lines) => {
@@ -51,7 +52,7 @@ impl LineGroup {
                     return vec![];
                 }
 
-                for line in lhs_lines.start.0..=lhs_lines.end.0 {
+                for line in lhs_lines.start.0..lhs_lines.end.0 {
                     res.push(line.into());
                 }
             }
@@ -60,18 +61,15 @@ impl LineGroup {
         res
     }
 
-    fn rhs_lines(&self) -> Vec<LineNumber> {
+    pub fn rhs_lines(&self) -> Vec<LineNumber> {
         let mut res = vec![];
         match &self.rhs_lines {
             Some(rhs_lines) => {
-                // TODO: it's not possible to represent an empty range
-                // with RangeInclusive and LineNumber. Define a
-                // separate interval type.
                 if rhs_lines.end.0 == 0 {
                     return vec![];
                 }
 
-                for line in rhs_lines.start.0..=rhs_lines.end.0 {
+                for line in rhs_lines.start.0..rhs_lines.end.0 {
                     res.push(line.into());
                 }
             }
@@ -83,14 +81,14 @@ impl LineGroup {
     pub fn pad(&mut self, amount: usize, max_lhs_line: LineNumber, max_rhs_line: LineNumber) {
         if let Some(Interval { start, end }) = self.lhs_lines.take() {
             let start = (max(0, start.0 as isize - amount as isize) as usize).into();
-            let end = min(max_lhs_line.0, end.0 + amount).into();
+            let end = min(max_lhs_line.0 + 1, end.0 + amount).into();
 
             self.lhs_lines = Some(Interval { start, end });
         }
 
         if let Some(Interval { start, end }) = self.rhs_lines.take() {
             let start = (max(0, start.0 as isize - amount as isize) as usize).into();
-            let end = min(max_rhs_line.0, end.0 + amount).into();
+            let end = min(max_rhs_line.0 + 1, end.0 + amount).into();
 
             self.rhs_lines = Some(Interval { start, end });
         }
@@ -277,12 +275,8 @@ pub fn visible_groups(
     lhs_positions: &[MatchedPos],
     rhs_positions: &[MatchedPos],
 ) -> Vec<LineGroup> {
-    let lhs_positions = lhs_positions
-        .iter()
-        .filter(|mp| mp.kind != MatchKind::Unchanged);
-    let rhs_positions = rhs_positions
-        .iter()
-        .filter(|mp| mp.kind != MatchKind::Unchanged);
+    let lhs_positions = lhs_positions.iter().filter(|mp| !mp.kind.is_unchanged());
+    let rhs_positions = rhs_positions.iter().filter(|mp| !mp.kind.is_unchanged());
 
     let mut positions = vec![];
     positions.extend(lhs_positions.map(|p| (true, p)));
@@ -321,141 +315,8 @@ pub fn visible_groups(
     groups
 }
 
-#[test]
-fn test_visible_groups_ignores_unchanged() {
-    let lhs_positions = vec![MatchedPos {
-        kind: MatchKind::Unchanged,
-        pos: vec![SingleLineSpan {
-            line: 1.into(),
-            start_col: 0,
-            end_col: 1,
-        }],
-        prev_opposite_pos: vec![],
-    }];
-    let rhs_positions = vec![MatchedPos {
-        kind: MatchKind::Unchanged,
-        pos: vec![SingleLineSpan {
-            line: 1.into(),
-            start_col: 0,
-            end_col: 1,
-        }],
-        prev_opposite_pos: vec![],
-    }];
-
-    let res = visible_groups(&lhs_positions, &rhs_positions);
-    assert_eq!(res, vec![]);
-}
-
-pub fn format_line_num(line_num: usize) -> String {
-    format!("{:<2} ", line_num + 1)
-}
-
-pub fn format_line_num_padded(line_num: usize, column_width: usize) -> String {
-    format!("{:width$} ", line_num + 1, width = column_width - 1)
-}
-
-pub fn lhs_printable_width(lhs: &str, lhs_column_width: usize, terminal_width: usize) -> usize {
-    let longest_line_length = lhs.lines().map(|line| line.len()).max().unwrap_or(1);
-    let longest_line = longest_line_length + lhs_column_width;
-
-    let space_available = terminal_width / 2 - SPACER.len();
-    max(MIN_WIDTH, min(longest_line, space_available))
-}
-
-pub fn rhs_printable_width(
-    rhs: &str,
-    lhs_width: usize,
-    rhs_column_width: usize,
-    terminal_width: usize,
-) -> usize {
-    let longest_line_length = rhs.lines().map(|line| line.len()).max().unwrap_or(1);
-    let longest_line = longest_line_length + rhs_column_width;
-
-    let space_available = (terminal_width - SPACER.len()) - lhs_width;
-
-    max(MIN_WIDTH, min(longest_line, space_available))
-}
-
-fn apply_group(
-    lhs_lines: &[&str],
-    rhs_lines: &[&str],
-    group: &LineGroup,
-    lhs_line_matches: &HashMap<LineNumber, LineNumber>,
-    lhs_content_width: usize,
-    lhs_column_width: usize,
-    rhs_column_width: usize,
-) -> String {
-    let mut result = String::new();
-
-    for (lhs_line_num, rhs_line_num) in
-        aligned_lines(&group.lhs_lines(), &group.rhs_lines(), lhs_line_matches)
-    {
-        match lhs_line_num {
-            Some(lhs_line_num) => {
-                result.push_str(&format_line_num_padded(lhs_line_num.0, lhs_column_width));
-                result.push_str(lhs_lines[lhs_line_num.0]);
-            }
-            None => {
-                result.push_str(&" ".repeat(lhs_column_width));
-                result.push_str(&" ".repeat(lhs_content_width));
-            }
-        }
-        result.push_str(SPACER);
-
-        match rhs_line_num {
-            Some(rhs_line_num) => {
-                result.push_str(&format_line_num_padded(rhs_line_num.0, rhs_column_width));
-                result.push_str(rhs_lines[rhs_line_num.0]);
-            }
-            None => {}
-        }
-
-        result.push('\n');
-    }
-
-    result
-}
-
-/// Display all the lines in `lhs` and `rhs` that are mentioned in
-/// `groups`. horizontally concatenating the matched lines.
-pub fn apply_groups(
-    lhs: &str,
-    rhs: &str,
-    groups: &[LineGroup],
-    lhs_line_matches: &HashMap<LineNumber, LineNumber>,
-    lhs_content_width: usize,
-    rhs_content_width: usize,
-    lhs_column_width: usize,
-    rhs_column_width: usize,
-) -> String {
-    let lhs_lines: Vec<_> = lhs.lines().collect();
-    let rhs_lines: Vec<_> = rhs.lines().collect();
-
-    let mut result = String::new();
-
-    let mut spacer = String::new();
-    spacer.push_str(&" ".repeat(lhs_column_width));
-    spacer.push_str(&"-".repeat(lhs_content_width));
-    spacer.push_str(&" ".repeat(rhs_column_width + 2));
-    spacer.push_str(&"-".repeat(rhs_content_width));
-
-    for (i, group) in groups.iter().enumerate() {
-        result.push_str(&apply_group(
-            &lhs_lines,
-            &rhs_lines,
-            group,
-            lhs_line_matches,
-            lhs_content_width,
-            lhs_column_width,
-            rhs_column_width,
-        ));
-        if i != groups.len() - 1 {
-            result.push_str(&spacer);
-            result.push('\n');
-        }
-    }
-
-    result
+pub fn format_line_num(line_num: LineNumber) -> String {
+    format!("{:>2} ", line_num.0 + 1)
 }
 
 /// A position in a single line of a string.
@@ -505,7 +366,7 @@ impl NewlinePositions {
             if region_start > line_end {
                 continue;
             }
-            if *line_start >= region_end {
+            if *line_start > region_end {
                 break;
             }
 
@@ -553,63 +414,6 @@ impl NewlinePositions {
     }
 }
 
-#[test]
-fn from_ranges_first_line() {
-    let newline_positions: NewlinePositions = "foo".into();
-    let line_spans = newline_positions.from_offsets(1, 3);
-    assert_eq!(
-        line_spans,
-        vec![SingleLineSpan {
-            line: 0.into(),
-            start_col: 1,
-            end_col: 3
-        }]
-    );
-}
-
-#[test]
-fn from_ranges_split_over_multiple_lines() {
-    let newline_positions: NewlinePositions = "foo\nbar\nbaz\naaaaaaaaaaa".into();
-    let line_spans = newline_positions.from_offsets(5, 10);
-
-    assert_eq!(
-        line_spans,
-        vec![
-            (SingleLineSpan {
-                line: 1.into(),
-                start_col: 1,
-                end_col: 3
-            }),
-            (SingleLineSpan {
-                line: 2.into(),
-                start_col: 0,
-                end_col: 2
-            })
-        ]
-    );
-}
-
-#[test]
-fn from_offsets_relative_to() {
-    let newline_positions: NewlinePositions = "foo\nbar".into();
-
-    let pos = SingleLineSpan {
-        line: 1.into(),
-        start_col: 1,
-        end_col: 1,
-    };
-
-    let line_spans = newline_positions.from_offsets_relative_to(pos, 1, 2);
-    assert_eq!(
-        line_spans,
-        vec![SingleLineSpan {
-            line: 1.into(),
-            start_col: 2,
-            end_col: 3
-        }]
-    );
-}
-
 /// Ensure that every line in `s` has this length. Pad short lines and
 /// truncate long lines.
 pub fn enforce_length(s: &str, line_length: usize) -> String {
@@ -629,16 +433,21 @@ pub fn enforce_length(s: &str, line_length: usize) -> String {
     result
 }
 
-#[test]
-fn enforce_length_short() {
-    let result = enforce_length("foo\nbar\n", 5);
-    assert_eq!(result, "foo  \nbar  \n");
-}
+/// Truncate any lines in `s` that are longer than `line_length`.
+pub fn enforce_max_length(s: &str, line_length: usize) -> String {
+    let mut result = String::with_capacity(s.len());
+    for line in s.lines() {
+        // TODO: use length in chars not bytes.
+        if line.len() > line_length {
+            // Truncate.
+            result.push_str(&line[0..line_length]);
+            result.push('\n');
+        } else {
+            result.push_str(&format!("{}\n", line));
+        }
+    }
 
-#[test]
-fn enforce_length_long() {
-    let result = enforce_length("foobar\nbarbaz\n", 3);
-    assert_eq!(result, "foo\nbar\n");
+    result
 }
 
 pub trait MaxLine {
@@ -651,14 +460,147 @@ impl MaxLine for String {
     }
 }
 
-#[test]
-fn str_max_line() {
-    let line: String = "foo\nbar".into();
-    assert_eq!(line.max_line().0, 1);
-}
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use crate::syntax::MatchKind;
+    use pretty_assertions::assert_eq;
 
-#[test]
-fn empty_str_max_line() {
-    let line: String = "".into();
-    assert_eq!(line.max_line().0, 0);
+    #[test]
+    fn from_ranges_first_line() {
+        let newline_positions: NewlinePositions = "foo".into();
+        let line_spans = newline_positions.from_offsets(1, 3);
+        assert_eq!(
+            line_spans,
+            vec![SingleLineSpan {
+                line: 0.into(),
+                start_col: 1,
+                end_col: 3
+            }]
+        );
+    }
+
+    #[test]
+    fn from_ranges_first_char() {
+        let newline_positions: NewlinePositions = "foo".into();
+        let line_spans = newline_positions.from_offsets(0, 0);
+        assert_eq!(
+            line_spans,
+            vec![SingleLineSpan {
+                line: 0.into(),
+                start_col: 0,
+                end_col: 0
+            }]
+        );
+    }
+
+    #[test]
+    fn test_visible_groups_ignores_unchanged() {
+        let lhs_positions = vec![MatchedPos {
+            kind: MatchKind::Unchanged {
+                opposite_pos: (vec![], vec![]),
+            },
+            pos: vec![SingleLineSpan {
+                line: 1.into(),
+                start_col: 0,
+                end_col: 1,
+            }],
+            prev_opposite_pos: vec![],
+        }];
+        let rhs_positions = vec![MatchedPos {
+            kind: MatchKind::Unchanged {
+                opposite_pos: (vec![], vec![]),
+            },
+            pos: vec![SingleLineSpan {
+                line: 1.into(),
+                start_col: 0,
+                end_col: 1,
+            }],
+            prev_opposite_pos: vec![],
+        }];
+
+        let res = visible_groups(&lhs_positions, &rhs_positions);
+        assert_eq!(res, vec![]);
+    }
+
+    #[test]
+    fn from_ranges_split_over_multiple_lines() {
+        let newline_positions: NewlinePositions = "foo\nbar\nbaz\naaaaaaaaaaa".into();
+        let line_spans = newline_positions.from_offsets(5, 10);
+
+        assert_eq!(
+            line_spans,
+            vec![
+                (SingleLineSpan {
+                    line: 1.into(),
+                    start_col: 1,
+                    end_col: 3
+                }),
+                (SingleLineSpan {
+                    line: 2.into(),
+                    start_col: 0,
+                    end_col: 2
+                })
+            ]
+        );
+    }
+
+    #[test]
+    fn enforce_length_short() {
+        let result = enforce_length("foo\nbar\n", 5);
+        assert_eq!(result, "foo  \nbar  \n");
+    }
+
+    #[test]
+    fn enforce_length_long() {
+        let result = enforce_length("foobar\nbarbaz\n", 3);
+        assert_eq!(result, "foo\nbar\n");
+    }
+
+    #[test]
+    fn str_max_line() {
+        let line: String = "foo\nbar".into();
+        assert_eq!(line.max_line().0, 1);
+    }
+
+    #[test]
+    fn empty_str_max_line() {
+        let line: String = "".into();
+        assert_eq!(line.max_line().0, 0);
+    }
+
+    #[test]
+    fn from_offsets_relative_to() {
+        let newline_positions: NewlinePositions = "foo\nbar".into();
+
+        let pos = SingleLineSpan {
+            line: 1.into(),
+            start_col: 1,
+            end_col: 1,
+        };
+
+        let line_spans = newline_positions.from_offsets_relative_to(pos, 1, 2);
+        assert_eq!(
+            line_spans,
+            vec![SingleLineSpan {
+                line: 1.into(),
+                start_col: 2,
+                end_col: 3
+            }]
+        );
+    }
+
+    #[test]
+    fn line_group_uses_interval_correctly() {
+        let group = LineGroup {
+            lhs_lines: Some(Interval {
+                start: 1.into(),
+                end: 3.into(),
+            }),
+            rhs_lines: None,
+        };
+
+        // Intervals are inclusive of `start` but exclusive of `end`.
+        assert_eq!(group.lhs_lines(), vec![1.into(), 2.into()])
+    }
 }
