@@ -1,3 +1,5 @@
+//! Apply colours and styling to strings.
+
 use crate::lines::LineNumber;
 use crate::positions::SingleLineSpan;
 use crate::syntax::{MatchKind, MatchedPos};
@@ -6,7 +8,7 @@ use std::cmp::min;
 use std::collections::HashMap;
 
 #[derive(Clone, Copy)]
-pub struct Style {
+struct Style {
     foreground: Color,
     background: Option<Color>,
     bold: bool,
@@ -25,7 +27,7 @@ impl Style {
     }
 }
 
-pub fn apply_line(line: &str, styles: &[(SingleLineSpan, Style)]) -> String {
+fn apply_line(line: &str, styles: &[(SingleLineSpan, Style)]) -> String {
     if styles.is_empty() {
         return line.dimmed().to_string();
     }
@@ -70,7 +72,7 @@ fn group_by_line(
 /// doesn't have any styles applied.
 ///
 /// Tolerant against lines in `s` being shorter than the spans.
-pub fn apply(s: &str, styles: &[(SingleLineSpan, Style)]) -> String {
+fn apply(s: &str, styles: &[(SingleLineSpan, Style)]) -> String {
     let mut ranges_by_line = group_by_line(styles);
 
     let mut res = String::with_capacity(s.len());
@@ -86,7 +88,7 @@ pub fn apply_colors(s: &str, is_lhs: bool, positions: &[MatchedPos]) -> String {
     let mut styles = vec![];
     for pos in positions {
         let style = match pos.kind {
-            MatchKind::Unchanged => Style {
+            MatchKind::Unchanged { .. } => Style {
                 foreground: Color::White,
                 background: None,
                 bold: false,
@@ -100,7 +102,7 @@ pub fn apply_colors(s: &str, is_lhs: bool, positions: &[MatchedPos]) -> String {
                 background: None,
                 bold: true,
             },
-            MatchKind::UnchangedCommentPart => Style {
+            MatchKind::UnchangedCommentPart { .. } => Style {
                 foreground: if is_lhs { Color::Red } else { Color::Green },
                 background: None,
                 bold: false,
@@ -115,5 +117,5 @@ pub fn apply_colors(s: &str, is_lhs: bool, positions: &[MatchedPos]) -> String {
 }
 
 pub fn header(file_name: &str, language_name: &str) -> String {
-    format!("{} ({})", file_name.yellow().bold(), language_name)
+    format!("{} -- {}", file_name.yellow().bold(), language_name)
 }
