@@ -1,3 +1,5 @@
+//! Syntax tree definitions with change metadata.
+
 #![allow(clippy::mutable_key_type)] // Hash for Syntax doesn't use mutable fields.
 
 use itertools::{EitherOrBoth, Itertools};
@@ -11,7 +13,7 @@ use std::fmt;
 use std::hash::{Hash, Hasher};
 use typed_arena::Arena;
 
-use crate::lines::{LineNumber, NewlinePositions};
+use crate::lines::{LineGroup, LineNumber, NewlinePositions};
 use crate::positions::SingleLineSpan;
 use ChangeKind::*;
 use Syntax::*;
@@ -42,6 +44,7 @@ pub struct SyntaxInfo<'a> {
     // TODO: Make these fields private.
     pub pos_content_hash: u64,
     pub next: Cell<Option<&'a Syntax<'a>>>,
+    pub prev: Cell<Option<&'a Syntax<'a>>>,
     pub change: Cell<Option<ChangeKind<'a>>>,
     pub num_ancestors: Cell<u64>,
     pub unique_id: Cell<u64>,
@@ -52,6 +55,7 @@ impl<'a> SyntaxInfo<'a> {
         Self {
             pos_content_hash,
             next: Cell::new(None),
+            prev: Cell::new(None),
             change: Cell::new(None),
             num_ancestors: Cell::new(0),
             unique_id: Cell::new(0),
@@ -158,7 +162,6 @@ fn trim_left(max_trim: usize, content: &str, pos: SingleLineSpan) -> (String, Si
 }
 
 impl<'a> Syntax<'a> {
-    #[allow(clippy::mut_from_ref)] // Clippy doesn't understand arenas.
     pub fn new_list(
         arena: &'a Arena<Syntax<'a>>,
         open_content: &str,
@@ -166,7 +169,7 @@ impl<'a> Syntax<'a> {
         children: Vec<&'a Syntax<'a>>,
         close_content: &str,
         close_position: Vec<SingleLineSpan>,
-    ) -> &'a mut Syntax<'a> {
+    ) -> &'a Syntax<'a> {
         let mut num_descendants = 0;
         for child in &children {
             num_descendants += match child {
@@ -198,21 +201,19 @@ impl<'a> Syntax<'a> {
         })
     }
 
-    #[allow(clippy::mut_from_ref)] // Clippy doesn't understand arenas.
     pub fn new_atom(
         arena: &'a Arena<Syntax<'a>>,
         position: Vec<SingleLineSpan>,
         content: &str,
-    ) -> &'a mut Syntax<'a> {
+    ) -> &'a Syntax<'a> {
         Self::new_atom_(arena, position, content, false)
     }
 
-    #[allow(clippy::mut_from_ref)] // Clippy doesn't understand arenas.
     pub fn new_comment(
         arena: &'a Arena<Syntax<'a>>,
         position: Vec<SingleLineSpan>,
         content: &str,
-    ) -> &'a mut Syntax<'a> {
+    ) -> &'a Syntax<'a> {
         // Ignore leading whitespace in multiline comments, so changes
         // in comment indentation are ignored.
         let first_line_indent = match position.first() {
@@ -267,10 +268,37 @@ impl<'a> Syntax<'a> {
         self.info().next.get()
     }
 
+    pub fn prev_is_contiguous(&self) -> bool {
+        if let Some(prev) = self.info().prev.get() {
+            match prev {
+                List {
+                    open_position,
+                    close_position,
+                    ..
+                } => {
+                    let prev_is_parent = prev.num_ancestors() < self.num_ancestors();
+                    if prev_is_parent {
+                        open_position.last().map(|p| p.line) == self.first_line()
+                    } else {
+                        // predecessor node at the same level.
+                        close_position.last().map(|p| p.line) == self.first_line()
+                    }
+                }
+                Atom { .. } => prev.last_line() == self.first_line(),
+            }
+        } else {
+            false
+        }
+    }
+
     pub fn id(&self) -> u64 {
         self.info().unique_id.get()
     }
 
+    pub fn num_ancestors(&self) -> u64 {
+        self.info().num_ancestors.get()
+    }
+
     pub fn first_line(&self) -> Option<LineNumber> {
         let position = match self {
             List { open_position, .. } => open_position,
@@ -414,6 +442,7 @@ impl<'a> Syntax<'a> {
 pub fn init_info<'a>(roots: &[&'a Syntax<'a>]) {
     set_unique_id(roots, 0);
     set_next(roots, None);
+    set_prev(roots, None);
     set_num_ancestors(roots, 0);
 }
 
@@ -429,6 +458,8 @@ fn set_unique_id<'a>(nodes: &[&'a Syntax<'a>], prev_id: u64) -> u64 {
     id
 }
 
+/// For every syntax node in the tree, mark the next node according to
+/// a preorder traversal.
 fn set_next<'a>(nodes: &[&'a Syntax<'a>], parent_next: Option<&'a Syntax<'a>>) {
     for (i, node) in nodes.iter().enumerate() {
         let node_next = match nodes.get(i + 1) {
@@ -443,6 +474,19 @@ fn set_next<'a>(nodes: &[&'a Syntax<'a>], parent_next: Option<&'a Syntax<'a>>) {
     }
 }
 
+/// For every syntax node in the tree, mark the previous node
+/// according to a preorder traversal.
+fn set_prev<'a>(nodes: &[&'a Syntax<'a>], parent: Option<&'a Syntax<'a>>) {
+    for (i, node) in nodes.iter().enumerate() {
+        let node_prev = if i == 0 { parent } else { Some(nodes[i - 1]) };
+
+        node.info().prev.set(node_prev);
+        if let List { children, .. } = node {
+            set_prev(children, Some(node));
+        }
+    }
+}
+
 fn set_num_ancestors<'a>(nodes: &[&Syntax<'a>], num_ancestors: u64) {
     for node in nodes {
         node.info().num_ancestors.set(num_ancestors);
@@ -466,18 +510,33 @@ impl<'a> Hash for Syntax<'a> {
     }
 }
 
-#[derive(PartialEq, Eq, Debug)]
+#[derive(PartialEq, Eq, Debug, Clone)]
 pub enum MatchKind {
-    Unchanged,
+    Unchanged {
+        opposite_pos: (Vec<SingleLineSpan>, Vec<SingleLineSpan>),
+    },
     Novel,
-    UnchangedCommentPart,
+    UnchangedCommentPart {
+        opposite_pos: Vec<SingleLineSpan>,
+    },
     ChangedCommentPart,
 }
 
-#[derive(Debug)]
+impl MatchKind {
+    pub fn is_unchanged(&self) -> bool {
+        match self {
+            MatchKind::Unchanged { .. } => true,
+            _ => false,
+        }
+    }
+}
+
+#[derive(Debug, Clone)]
 pub struct MatchedPos {
     pub kind: MatchKind,
     pub pos: Vec<SingleLineSpan>,
+    // TODO: this is confusing: the previous syntax node with a match
+    // may be on the current line or a previous one.
     pub prev_opposite_pos: Vec<SingleLineSpan>,
 }
 
@@ -485,6 +544,7 @@ fn split_comment_words(
     content: &str,
     pos: &[SingleLineSpan],
     opposite_content: &str,
+    opposite_pos: &[SingleLineSpan],
     prev_opposite_pos: &[SingleLineSpan],
 ) -> Vec<MatchedPos> {
     // TODO: also split on whitespace, so "// (foo)" splits before "(".
@@ -498,8 +558,10 @@ fn split_comment_words(
     let other_parts: Vec<_> = WORD_BOUNDARY_RE.split(opposite_content).collect();
 
     let content_newlines = NewlinePositions::from(content);
+    let opposite_content_newlines = NewlinePositions::from(opposite_content);
 
     let mut offset = 0;
+    let mut opposite_offset = 0;
 
     let mut res = vec![];
     for diff_res in diff::slice(&content_parts, &other_parts) {
@@ -517,21 +579,29 @@ fn split_comment_words(
                 });
                 offset += word.len();
             }
-            diff::Result::Both(word, _) => {
+            diff::Result::Both(word, opposite_word) => {
                 // This word is present on both sides.
+                let word_pos =
+                    content_newlines.from_offsets_relative_to(pos[0], offset, offset + word.len());
+                let opposite_word_pos = opposite_content_newlines.from_offsets_relative_to(
+                    opposite_pos[0],
+                    opposite_offset,
+                    opposite_offset + opposite_word.len(),
+                );
+
                 res.push(MatchedPos {
-                    kind: MatchKind::UnchangedCommentPart,
-                    pos: content_newlines.from_offsets_relative_to(
-                        pos[0],
-                        offset,
-                        offset + word.len(),
-                    ),
+                    kind: MatchKind::UnchangedCommentPart {
+                        opposite_pos: opposite_word_pos,
+                    },
+                    pos: word_pos,
                     prev_opposite_pos: prev_opposite_pos.to_vec(),
                 });
                 offset += word.len();
+                opposite_offset += opposite_word.len();
             }
-            diff::Result::Right(_) => {
+            diff::Result::Right(opposite_word) => {
                 // Only exists on other side, nothing to do on this side.
+                opposite_offset += opposite_word.len();
             }
         }
     }
@@ -551,19 +621,33 @@ impl MatchedPos {
                     List { .. } => unreachable!(),
                     Atom { content, .. } => content,
                 };
-                let opposite_content = match opposite {
+                let (opposite_content, opposite_pos) = match opposite {
                     List { .. } => unreachable!(),
-                    Atom { content, .. } => content,
+                    Atom {
+                        content, position, ..
+                    } => (content, position),
                 };
 
                 return split_comment_words(
                     this_content,
                     &pos,
                     opposite_content,
+                    opposite_pos,
                     &prev_opposite_pos,
                 );
             }
-            Unchanged(_) => MatchKind::Unchanged,
+            Unchanged(opposite) => {
+                let opposite_pos = match opposite {
+                    List {
+                        open_position,
+                        close_position,
+                        ..
+                    } => (open_position.clone(), close_position.clone()),
+                    Atom { position, .. } => (position.clone(), position.clone()),
+                };
+
+                MatchKind::Unchanged { opposite_pos }
+            }
             Novel => MatchKind::Novel,
         };
 
@@ -673,7 +757,7 @@ fn change_positions_<'a>(
                     match opposite_node {
                         List { .. } => {
                             dbg!(node, opposite_node);
-                            unreachable!()
+                            unreachable!();
                         }
                         Atom {
                             position: opposite_position,
@@ -741,25 +825,18 @@ fn zip_lines(lhs: &[SingleLineSpan], rhs: &[SingleLineSpan]) -> Vec<(LineNumber,
 }
 
 pub fn aligned_lines(
-    lhs_lines: &[LineNumber],
-    rhs_lines: &[LineNumber],
+    group: &LineGroup,
     lhs_line_matches: &HashMap<LineNumber, LineNumber>,
 ) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
+    let lhs_lines = group.lhs_lines();
+    let rhs_lines = group.rhs_lines();
+
     // When adding padding to a LineGroup where each side has a
     // different number of lines, we can end up with extra padding on
     // the side with fewer lines.
     //
-    // Work around that by discarding any lines at the beginning with
-    // aren't matched.
-    //
     // TODO: fix padding to be smarter.
-    //
-    // TODO: do the same for the end of the hunk.
-    aligned_lines_(lhs_lines, rhs_lines, lhs_line_matches)
-        .iter()
-        .skip_while(|(lhs, rhs)| lhs.is_none() || rhs.is_none())
-        .copied()
-        .collect()
+    aligned_lines_(&lhs_lines, &rhs_lines, lhs_line_matches)
 }
 
 /// Given two slices of contiguous line numbers, return pairs of
@@ -775,13 +852,13 @@ fn aligned_lines_(
     rhs_lines: &[LineNumber],
     lhs_line_matches: &HashMap<LineNumber, LineNumber>,
 ) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    let mut rhs_highest_matched = rhs_lines.first().map(|l| l.0 as isize).unwrap_or(0) - 1;
+    let mut rhs_highest_matched = rhs_lines.first().map_or(0, |l| l.0 as isize) - 1;
 
     // For every LHS line, if there is a RHS line that is included in
     // `rhs_lines` and hasn't yet been paired up, add it to
     // matched_lines.
     //
-    // TODO: prefer the lie with the most similarity, not just the
+    // TODO: prefer the line with the most similarity, not just the
     // first line.  See spurious `let` alignment in 9c71298f8294ce8f,
     // LHS line 96 in lines.rs.
     let mut matched_lines = vec![];
@@ -903,20 +980,6 @@ mod tests {
         );
     }
 
-    #[test]
-    fn test_aligned_middle_drop_none_at_start() {
-        let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
-        let rhs_lines: Vec<LineNumber> = vec![12.into(), 13.into()];
-
-        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-        line_matches.insert(2.into(), 12.into());
-
-        assert_eq!(
-            aligned_lines(&lhs_lines, &rhs_lines, &line_matches),
-            vec![(Some(2.into()), Some(12.into())), (None, Some(13.into()))]
-        );
-    }
-
     #[test]
     fn test_aligned_all() {
         let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
@@ -1025,7 +1088,7 @@ mod tests {
                 start_col: 2,
                 end_col: 3,
             }],
-            "foo".into(),
+            "foo",
         );
         atom.set_change(ChangeKind::Novel);
         let nodes: Vec<&Syntax> = vec![atom];
@@ -1051,8 +1114,8 @@ mod tests {
 
         let arena = Arena::new();
 
-        let comment = Syntax::new_comment(&arena, pos.clone(), "foo".into());
-        let atom = Syntax::new_atom(&arena, pos, "foo".into());
+        let comment = Syntax::new_comment(&arena, pos.clone(), "foo");
+        let atom = Syntax::new_atom(&arena, pos, "foo");
 
         assert_ne!(comment, atom);
     }
